import os
from argparse import ArgumentParser

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import six
from dodo_commands.framework.command_map import get_command_map
from dodo_commands.framework.command_path import get_command_dirs_from_config

raw_input = six.moves.input


def _args():
    parser = ArgumentParser(description="Creates a new Dodo command.")
    parser.add_argument("name")
    parser.add_argument("--next-to")
    parser.add_argument(
        "-f", "--force", action="store_true", help="Overwrite existing command script"
    )

    args = Dodo.parse_args(parser)
    return args


script_py = """from argparse import ArgumentParser

from dodo_commands import Dodo, ConfigArg, CommandError


def _args():
    # Create the parser
    parser = ArgumentParser(description='')
    parser.add_argument('foo')

    # Use the parser to create the command arguments
    args = Dodo.parse_args(parser, config_args=[])
    args.cwd = Dodo.get('/ROOT/project_dir')

    # Raise an error if something is not right
    if False:
        raise CommandError('Oops')

    return args


# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()
    Dodo.run(['echo', args.foo], cwd=args.cwd)
"""


def handle_next_to(args):
    command_dirs = get_command_dirs_from_config(Dodo.get())
    command_map = get_command_map(command_dirs)

    if args.next_to:
        item = command_map.get(args.next_to)
        if not item:
            raise CommandError("Script not found: %s" % args.next_to)
        dest_path = os.path.join(os.path.dirname(item.filename), args.name + ".py")

        if os.path.exists(dest_path) and not args.force:
            raise CommandError("Destination already exists: %s" % dest_path)

        with open(dest_path, "w") as f:
            f.write(
                script_py.format(
                    parser_args_str="", params_str="", description="", args_str=""
                )
            )
        print(dest_path)
    else:
        print(
            script_py.format(
                parser_args_str="", params_str="", description="", args_str=""
            )
        )


if Dodo.is_main(__name__, safe=False):
    args = _args()
    handle_next_to(args)



"""Decorates command lines with docker arguments."""

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.args_tree import ArgsTreeNode

local = plumbum.local


class Decorator:  # noqa
    @classmethod
    def merged_options(cls, get_config):
        return get_config("/KUBERNETES", {})

    @classmethod
    def kubectl_node(cls, get_config):
        full_container_name = cls.get_full_container_name(get_config)

        if full_container_name:
            return ArgsTreeNode(
                "kubectl",
                args=["kubectl", "exec", "-it", full_container_name, "--"],
            )
        else:
            raise CommandError("No kubernetes container found")

    @classmethod
    def get_full_container_name(cls, get_config):
        options = cls.merged_options(get_config)
        labels = [
            f"-l={key}={value}" for key, value in options.get("labels", {}).items()
        ]
        kubectl = local["kubectl"]

        if labels:
            full_container_name = kubectl(
                "get",
                "pods",
                *labels,
                "-o",
                'jsonpath="{.items[*].metadata.name}"',
                "--field-selector",
                "status.phase=Running",
            )[1:-1]
        else:
            full_container_name = None
        return full_container_name

    def add_arguments(self, parser):  # noqa
        pass

    def modify_args(self, command_line_args, args_tree_root_node, cwd):
        kubectl_node = self.kubectl_node(Dodo.get_config)

        kubectl_node.add_child(args_tree_root_node)
        return kubectl_node, local.cwd



from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.config import build_config, extend_command_path
from dodo_commands.framework.config_layers import get_conflicts_in_layer_paths


def check_conflicts_in_selected_layer_paths(ctr):
    conflicts = get_conflicts_in_layer_paths(ctr.layers.selected_layer_by_path.keys())
    for path, conflicting_path in conflicts:
        raise CommandError("Conflicting layers: %s and %s" % (path, conflicting_path))


def build_from_selected_layers(ctr):
    def get_selected_layers():
        return ctr.layers.selected_layer_by_path.values()

    x = get_selected_layers()
    config, warnings = build_config(x)
    config = extend_command_path(config)

    ctr.config.config = config
    ctr.config.warnings = warnings



"""Pauses the execution."""
import time


class Decorator:
    def add_arguments(self, parser):  # override
        parser.add_argument(
            "--pause-ms", type=int, help="Pause in milliseconds before continuing"
        )

    def modify_args(self, command_line_args, args_tree_root_node, cwd):  # override
        if getattr(command_line_args, "pause_ms", 0):
            time.sleep(command_line_args.pause_ms / 1000)
        return args_tree_root_node, cwd



import argparse
import os
import shlex

from dodo_commands.framework import ramda as R
from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.handle_arg_complete import handle_arg_complete


def parse_input_args(ctr):
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("-L", "--layer", action="append")
    parser.add_argument("-D", "--decorator", action="append")
    parser.add_argument("--env", action="append")
    parser.add_argument("--trace", action="store_true")
    parser.add_argument("--help", action="store_true")
    parser.add_argument("--cwd")

    known_args, args = parser.parse_known_args(ctr.command_line.input_args)

    ctr.command_line.layer_paths_from_input_args = known_args.layer or []
    ctr.command_line.decorators_from_input_args = known_args.decorator or []
    ctr.command_line.env_vars_from_input_args = known_args.env or []
    ctr.command_line.cwd = known_args.cwd or None
    ctr.command_line.is_trace = known_args.trace
    ctr.command_line.is_help = known_args.help
    ctr.command_line.input_args = args


def get_layer_paths_from_command_prefix(ctr):
    def map_to_path(layer_name):
        if layer_name not in ctr.layers.metadata_by_layer_name:
            known_layer_names = ctr.layers.metadata_by_layer_name.keys()
            raise CommandError(
                "Unknown layer: %s. Known layers: %s"
                % (layer_name, ", ".join(known_layer_names))
            )
        return ctr.layers.metadata_by_layer_name[layer_name].target_path

    layer_names = R.uniq(R.filter(bool)(ctr.command_line.command_prefix.split(".")))
    layer_paths_from_command_prefix = (
        R.map(map_to_path)(layer_names)
        or ctr.command_line.layer_paths_from_command_prefix
    )

    ctr.command_line.layer_paths_from_command_prefix = layer_paths_from_command_prefix


def expand_and_autocomplete_command_name(ctr):
    is_autocompleting = "_ARGCOMPLETE" in os.environ
    completed_command_name = (
        handle_arg_complete(
            command_names=list(ctr.commands.command_map.keys()),
            command_aliases=list(ctr.commands.aliases.keys()),
            metadata_by_layer_name=ctr.layers.metadata_by_layer_name,
            layer_by_target_path=ctr.layers.layer_by_target_path,
        )
        if is_autocompleting
        else ctr.command_line.command_name
    )

    alias_target = ctr.commands.aliases.get(completed_command_name)
    new_args = (
        shlex.split(alias_target) if alias_target else ctr.command_line.input_args[1:2]
    )

    ctr.command_line.input_args = (
        ctr.command_line.input_args[:1] + new_args + ctr.command_line.input_args[2:]
    )
    return not is_autocompleting and new_args != ctr.command_line.input_args[1:2]



from dodo_commands.framework.command_map import get_command_map as _get_command_map
from dodo_commands.framework.command_path import get_command_dirs_from_config
from dodo_commands.framework.get_aliases import get_aliases


def get_aliases_from_config(ctr):
    ctr.commands.aliases_from_config = get_aliases(ctr.config.config)


def get_command_map(ctr):
    command_dirs = get_command_dirs_from_config(ctr.config.config)

    ctr.commands.command_dirs = command_dirs
    ctr.commands.command_map = _get_command_map(command_dirs)



import os

from dodo_commands.dependencies import (yaml_round_trip_dump,
                                        yaml_round_trip_load)
from dodo_commands.framework.config import merge_into_config
from dodo_commands.framework.paths import Paths


def create_config_dir(config_dir):
    """Install the dir with dodo_commands resources."""
    os.makedirs(config_dir)
    config_filename = os.path.join(config_dir, "config.yaml")
    default_config = {
        "ROOT": {
            "command_path": [
                os.path.join(Paths().default_commands_dir(expanduser=False), "*")
            ],
            "version": "1.0.0",
        }
    }

    default_config_mixin_filename = Paths().default_config_mixin_filename()
    if os.path.exists(default_config_mixin_filename):
        with open(default_config_mixin_filename) as f:
            default_config_mixin = yaml_round_trip_load(f.read())
            merge_into_config(default_config, default_config_mixin)

    with open(config_filename, "w") as f:
        for key in default_config:
            f.write(yaml_round_trip_dump({key: default_config[key]}))
            f.write(os.linesep)

    os.makedirs(os.path.join(config_dir, ".dodo-start-env"))



from dodo_commands import Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.config import expand_keys
from dodo_commands.framework.global_config import load_global_config_parser
from dodo_commands.framework.util import to_arg_list


def _args():
    Dodo.parser.description = "Run docker compose"

    Dodo.parser.add_argument("compose_args", nargs="*")

    group = Dodo.parser.add_mutually_exclusive_group()
    group.add_argument("--cat", action="store_true")
    group.add_argument("--edit", action="store_true")
    args = Dodo.parse_args()

    key = "DOCKER_COMPOSE"
    args.cwd = Dodo.get("/" + key + "/cwd")
    args.progress = Dodo.get("/" + key + "/progress", "plain")
    args.files = Dodo.get("/" + key + "/files", None)
    args.env_file = Dodo.get("/" + key + "/env_file", None)
    args.map = Dodo.get("/" + key + "/map", {})
    args.editor = load_global_config_parser().get("settings", "editor")
    args.compose_project_name = Dodo.get(
        "/" + key + "/compose_project_name", Dodo.get("/ROOT/env_name")
    )
    return args


# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()

    for src, dest in args.map.items():
        with open(src) as ifs:
            content = expand_keys(Dodo.get(), ifs.read())
        with open(dest, "w") as ofs:
            ofs.write(content)

    def get_file_args():
        result = []
        for f in args.files:
            result.extend(["--file", f])
        return result

    file_args = get_file_args() if args.files else []
    env_file_args = [f"--env-file={args.env_file}"] if args.env_file else []

    compose_args = to_arg_list(args.compose_args)
    if "build" in compose_args[:1]:
        if "--progress" not in compose_args:
            compose_args = [
                compose_args[0],
                "--progress",
                args.progress,
                *compose_args[1:],
            ]
    if args.cat:
        for f in args.files:
            Dodo.run(["cat", f], cwd=args.cwd)
    elif args.edit:
        Dodo.run([args.editor, *args.files], cwd=args.cwd)
    else:
        with plumbum.local.env(COMPOSE_PROJECT_NAME=args.compose_project_name):
            Dodo.run(
                [
                    "docker-compose",
                    *file_args,
                    *env_file_args,
                    *compose_args,
                ],
                cwd=args.cwd,
            )



from dodo_commands import Dodo
from dodo_commands.framework.util import to_arg_list


def _args():
    Dodo.parser.description = "Run something on the commandline"

    Dodo.parser.add_argument("--cwd")
    Dodo.parser.add_argument("script_args")

    args = Dodo.parse_args()
    return args


# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()
    Dodo.run(to_arg_list(args.script_args), cwd=args.cwd or None)



import sys

from dodo_commands.framework.container import facets
from dodo_commands.framework.container.actions.command_line import (
    expand_and_autocomplete_command_name,
    get_layer_paths_from_command_prefix,
    parse_input_args,
)
from dodo_commands.framework.container.actions.commands import (
    get_aliases_from_config,
    get_command_map,
)
from dodo_commands.framework.container.actions.config import (
    build_from_selected_layers,
    check_conflicts_in_selected_layer_paths,
)
from dodo_commands.framework.container.actions.layers import (
    get_metadata_by_layer_name,
    load_named_layers,
    load_root_layer,
    select_layers,
)
from dodo_commands.framework.paths import Paths


class Container:
    def _create_facets(self):
        self.paths = Paths()
        self.command_line = facets.CommandLine()
        self.layers = facets.init_layers(facets.Layers(), "config.yaml")
        self.config = facets.Config()
        self.commands = facets.init_commands(facets.Commands())

    def __init__(self):
        self._create_facets()

    @staticmethod
    def get(ctr):
        return ctr

    def run_actions(self):
        load_root_layer(self)
        get_metadata_by_layer_name(self)
        load_named_layers(self)
        parse_input_args(self)

        while True:
            get_layer_paths_from_command_prefix(self)

            # Create the config
            select_layers(self)
            check_conflicts_in_selected_layer_paths(self)
            build_from_selected_layers(self)

            get_aliases_from_config(self)
            get_command_map(self)

            if not expand_and_autocomplete_command_name(self):
                break

        if self.config.warnings:
            for warning in self.config.warnings:
                sys.stderr.write(warning)



from collections import OrderedDict

from dodo_commands.framework import ramda as R
from dodo_commands.framework.config import build_config
from dodo_commands.framework.config_layers import layer_filename_superset
from dodo_commands.framework.get_metadata_by_layer_name import (
    get_metadata_by_layer_name as get_metadata_by_layer_name_,
)
from dodo_commands.framework.load_named_layers import (
    load_named_layers as load_named_layers_,
)


def load_root_layer(ctr):
    root_layer = (
        ctr.layers.config_io.load(ctr.layers.root_layer_path)
        if ctr.paths.project_dir()
        else {}
    )
    root_layer_config = build_config([root_layer])[0]

    ctr.layers.root_layer = root_layer
    ctr.layers.root_layer_config = root_layer_config


def get_metadata_by_layer_name(ctr):
    ctr.layers.metadata_by_layer_name = get_metadata_by_layer_name_(
        ctr.layers.root_layer
    )


def load_named_layers(ctr):
    layer_by_target_path = load_named_layers_(
        ctr.layers.config_io,
        ctr.layers.root_layer_config,
        ctr.layers.metadata_by_layer_name,
    )

    ctr.layers.layer_by_target_path = layer_by_target_path


def select_layers(ctr):
    all_layer_paths = layer_filename_superset(
        R.uniq([ctr.layers.root_layer_path] + ctr.command_line.target_paths),
        config_io=ctr.layers.config_io,
    )

    selected_layer_by_path = OrderedDict()
    for layer_path in all_layer_paths:
        selected_layer_by_path[layer_path] = ctr.layers.config_io.load(layer_path)

    ctr.layers.selected_layer_by_path = selected_layer_by_path



import sys

from dodo_commands import Dodo
from dodo_commands.framework.util import maybe_list_to_list


def _args():
    Dodo.parser.add_argument("command")
    Dodo.parser.add_argument("run_args", nargs="*")
    args = Dodo.parse_args()

    return args


# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()

    commands = Dodo.get("/COMMANDS")
    if not commands:
        sys.exit(0)

    if args.command not in commands:
        for alias, command in Dodo.get_container().commands.aliases.items():
            if args.command == alias:
                args.command = command
                break

    command = commands[args.command]

    cmd = None
    cmd_args = []
    decorator_names = []
    for part in command.get("run", {}):
        # check if part is a dict
        if isinstance(part, dict) and "in" in part:
            for decorator_name in maybe_list_to_list(part["in"]):
                decorator_names.append(decorator_name)
        elif isinstance(part, dict) and "cmd" in part:
            cmd = part["cmd"]
        elif not isinstance(part, dict):
            cmd_args.append(part)
        else:
            raise Exception("Unknown part in command")
    if not cmd:
        raise Exception("No cmd found in command")
    is_dodo = cmd == "dodo"

    more_kwargs = {}
    more_args = []
    for arg in args.run_args:
        if arg.startswith("++"):
            if "=" in arg:
                k, v = arg.split("=")
                more_kwargs[k] = v
            else:
                more_args.append(arg)
        elif arg.startswith("+"):
            if "=" in arg:
                k, v = arg.split("=")
                more_kwargs[k] = v
            else:
                more_args.append(arg)
        else:
            more_args.append(arg)

    kwargs = command.get("kwargs", {})
    for kwarg, value in kwargs.items():
        if kwarg not in more_kwargs:
            cmd_args.append(f"{kwarg}={value}")
    for kwarg, value in more_kwargs.items():
        cmd_args.append(f"{kwarg}={value}")
    for arg in more_args:
        cmd_args.append(arg)

    cwd = command.get("cwd")

    if env := command.get("env"):
        if isinstance(env, dict):
            env = dict(env)
        else:
            variable_map = {}
            for env_var in env:
                key, value = env_var.split("=")
                variable_map[key] = value
            env = variable_map
        for key, value in env.items():
            if is_dodo:
                cmd_args.append(f"--env={key}={value}")
            else:
                for decorator_name in decorator_names:
                    Dodo.get_container().command_line.env_variables_from_input_args.append(
                        f"{key}={value}"
                    )

    quiet = False
    if is_dodo:
        quiet = True
        if cwd:
            cmd_args.append(f"--cwd={cwd}")
            cwd = None
        if Dodo._args.confirm:
            Dodo._args.confirm = 0
            cmd_args.append("--confirm")
        for decorator_name in decorator_names:
            cmd_args.append(f"--decorator={decorator_name}")
    else:

        def update_cmd_arg(arg):
            if arg.startswith("++"):
                return f"--{arg[2:]}"
            elif arg.startswith("+"):
                return f"-{arg[1:]}"
            return arg

        cmd_args = [update_cmd_arg(arg) for arg in cmd_args]
        for decorator_name in decorator_names:
            Dodo.get_container().command_line.decorators_from_input_args.append(
                decorator_name
            )

    Dodo.run([cmd, *cmd_args], quiet=quiet, cwd=cwd)



import argparse
import os
import sys

from dodo_commands.dependencies.get import argcomplete, plumbum
from dodo_commands.framework.args_tree import ArgsTreeNode
from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.config_arg import add_config_args
from dodo_commands.framework.config_key import Key, KeyNotFound
from dodo_commands.framework.container.container import Container
from dodo_commands.framework.decorator_utils import get_decorators
from dodo_commands.framework.util import classproperty


class Dodo:
    safe = True
    parser = argparse.ArgumentParser()

    _args = None
    _container = None

    @classmethod
    def get_container(cls):
        if cls._container is None:
            cls._container = Container()
            cls._container.run_actions()

        return cls._container

    @classproperty
    def command_name(cls):  # noqa
        return cls.get_container().command_line.command_name

    @classproperty
    def args(cls):  # noqa
        if cls._args is None:
            cls.parse_args(cls.parser)
        return cls._args

    @classproperty
    def command_name(cls):  # noqa
        return cls.get_container().command_line.command_name

    @classmethod
    def get(cls, key="", default_value="__not_set_234234__"):  # noqa
        try:
            return Key(cls.get_container().config.config, key).get()
        except KeyNotFound:
            if default_value == "__not_set_234234__":
                raise
            return default_value

    @classmethod
    def is_main(cls, name, safe=None):
        if safe is not None:
            cls.safe = safe

        for decorator in get_decorators(
            cls.command_name,
            cls.get(),
            decorators_from_input_args=(
                cls.get_container().command_line.decorators_from_input_args
            ),
        ):
            if hasattr(decorator, "install"):
                decorator.install()

        if name == "__main__":
            return True

        commands = cls.get_container().commands
        command_map_item = commands.command_map.get(cls.command_name, None)
        if not command_map_item:
            return False

        script_path = command_map_item.package_path + "." + cls.command_name
        return name == script_path

    @classmethod
    def parse_args(cls, parser=None, config_args=None):
        parser = parser or cls.parser

        parser.add_argument("--traceback", action="store_true", help=argparse.SUPPRESS)

        add_config_args(parser, cls.get(), config_args)

        for decorator in get_decorators(
            cls.command_name,
            cls.get(),
            decorators_from_input_args=(
                cls.get_container().command_line.decorators_from_input_args
            ),
        ):
            decorator.add_arguments(parser)

        args = cls.get_container().command_line.input_args
        parser.prog = os.path.basename(args[0])
        if len(args) > 1:
            parser.prog += " %s" % args[1]

        if cls.get_container().command_line.is_help:
            parser.print_help()
            sys.exit(0)

        argcomplete.autocomplete(parser)
        cls._args = parser.parse_args(args[2:])

        if config_args:
            for config_arg in config_args:
                if config_arg.exists(cls.get()):
                    setattr(cls._args, config_arg.arg_name, config_arg.get(cls.get()))

        return cls._args

    @classmethod
    def run(
        cls,
        args,
        cwd=None,
        quiet=False,
        capture=False,
    ):
        if cls.get_container().command_line.env_vars_from_input_args:
            for env_var in cls.get_container().command_line.env_vars_from_input_args:
                key, value = env_var.split("=")
                variable_map = (
                    Dodo.get()
                    .setdefault("ENVIRONMENT", {})
                    .setdefault("variable_map", {})
                )
                variable_map[key] = value

        cwd = cwd or cls.get_container().command_line.cwd
        args_tree_root_node = ArgsTreeNode("original_args", args=args)
        for decorator in get_decorators(
            cls.command_name,
            cls.get(),
            decorators_from_input_args=(
                cls.get_container().command_line.decorators_from_input_args
            ),
        ):
            args_tree_root_node, cwd = decorator.modify_args(
                cls._args, args_tree_root_node, cwd
            )
            if args_tree_root_node is None:
                return False

        if cwd and not os.path.exists(cwd):
            raise CommandError("Directory not found: %s" % cwd)

        with plumbum.local.cwd(cwd or plumbum.local.cwd):
            flat_args = args_tree_root_node.flatten()
            func = plumbum.local[flat_args[0]][flat_args[1:]]

            variable_map = cls.get("/ENVIRONMENT/variable_map", {})
            with plumbum.local.env(**variable_map):
                if capture:
                    return func()
                try:
                    func & plumbum.FG
                    return True
                except plumbum.ProcessExecutionError:
                    if not quiet:
                        print("\nDodo Commands error while running this command:")
                        print("\n\n%s" % func)
                    return False

    # add alias for the legacy 'runcmd' method
    runcmd = run
    get_config = get
